import 'package:flutter/material.dart';
import 'package:logging/logging.dart';
import 'package:media_kit/media_kit.dart';
import 'package:media_kit_video/media_kit_video.dart';

class MediaKitAvControl extends StatefulWidget {
  final String videoUrl;
  final double? width;
  final double? height;
  final Function()? onGoFullScreen;

  const MediaKitAvControl({
    super.key,
    required this.videoUrl,
    required this.width,
    required this.height,
    this.onGoFullScreen,
  });

  @override
  State<MediaKitAvControl> createState() => _MediaKitAvControlState();
}

class _MediaKitAvControlState extends State<MediaKitAvControl> {
  static final _logger = Logger('$MediaKitAvControl');
  final Player player = Player(
    configuration: const PlayerConfiguration(
      logLevel: MPVLogLevel.warn,
    ),
  );
  VideoController? controller;
  var needToOpen = true;
  var playing = false;

  @override
  void initState() {
    super.initState();
    Future.microtask(() async {
      _logger.info('initializing');
      controller = await VideoController.create(player);
      _logger.info('initialized');
      if (context.mounted) {
        setState(() {});
      }
    });
  }

  @override
  void dispose() {
    Future.microtask(() async {
      await controller?.dispose();
      await player.dispose();
    });
    super.dispose();
  }

  @override
  void deactivate() {
    player.pause();
    playing = false;
    super.deactivate();
  }

  Future<void> toggleVideoPlay() async {
    _logger.fine('Toggling play on: ${widget.videoUrl}');
    if (needToOpen) {
      await player.open(Media(widget.videoUrl), play: false);
      needToOpen = false;
    }
    if (playing) {
      await player.pause();
      playing = false;
    } else {
      await player.play();
      playing = true;
    }
    setState(() {});
  }

  void resetPlay() async {
    if (playing) {
      await player.pause();
    }

    await player.seek(Duration.zero);

    if (playing) {
      playing = false;
      await player.play();
      playing = true;
    }

    setState(() {});
  }

  double? get height => widget.height == null ? null : widget.height! - 50;

  double? get width => widget.width;

  @override
  Widget build(BuildContext context) {
    _logger.finer(
        'Building MediaKit Control playing? $playing for ${widget.videoUrl}');
    if (controller == null) {
      return Container(
        width: widget.width,
        height: widget.height,
        color: Colors.black12,
        child: Column(
          mainAxisAlignment: MainAxisAlignment.center,
          crossAxisAlignment: CrossAxisAlignment.center,
          children: const [
            CircularProgressIndicator(),
          ],
        ),
      );
    }

    return GestureDetector(
      onTap: toggleVideoPlay,
      child: Column(
        mainAxisAlignment: MainAxisAlignment.center,
        children: [
          Expanded(
            child: Video(
              controller: controller,
              width: width,
              height: height,
            ),
          ),
          Row(
            mainAxisAlignment: MainAxisAlignment.center,
            children: [
              IconButton(
                icon: playing
                    ? const Icon(Icons.pause)
                    : const Icon(Icons.play_arrow),
                onPressed: toggleVideoPlay,
              ),
              IconButton(onPressed: resetPlay, icon: const Icon(Icons.replay)),
              if (widget.onGoFullScreen != null)
                IconButton(
                  onPressed: () async {
                    if (playing) {
                      await toggleVideoPlay();
                    }
                    await widget.onGoFullScreen!();
                  },
                  icon: const Icon(Icons.fullscreen),
                )
            ],
          )
        ],
      ),
    );
  }
}
